@@ -33,13 +33,14 @@ enum {
 
 static void kz_gesture_class_init   (KzGestureClass *class);
 static void kz_gesture_init         (KzGesture *gesture);
+static void kz_gesture_dispose      (GObject *object);
 
 /* KzGesture class signals */
 static void kz_gesture_stack_motion (KzGesture *gesture,
 				     KzGestureMotion motion);
 
 static void kz_gesture_reset        (KzGesture *gesture);
-static KzGestureDefTable *kz_gesture_table_search (KzGesture *gesture);
+static KzGestureItem *kz_gesture_search_matched_item (KzGesture *gesture);
 
 static GObjectClass *parent_class = NULL;
 static gint kz_gesture_signals[LAST_SIGNAL] = {0};
@@ -51,7 +52,10 @@ KZ_OBJECT_GET_TYPE(kz_gesture, "KzGesture", KzGesture,
 static void
 kz_gesture_class_init (KzGestureClass *class)
 {
+	GObjectClass *object_class;
+
 	parent_class = g_type_class_peek_parent (class);
+	object_class = (GObjectClass *) class;
 
 	kz_gesture_signals[STACK_MOTION_SIGNAL]
 		= g_signal_new ("stack-motion",
@@ -62,6 +66,8 @@ kz_gesture_class_init (KzGestureClass *class)
 				g_cclosure_marshal_VOID__ENUM,
 				G_TYPE_NONE, 1, G_TYPE_INT);
 
+	object_class->dispose = kz_gesture_dispose;
+
 	class->start        = NULL;
 	class->stack_motion = kz_gesture_stack_motion;
 	class->perform      = NULL;
@@ -71,13 +77,29 @@ static void
 kz_gesture_init (KzGesture *gesture)
 {
 	gesture->max_sequence_len
-		= sizeof(gesture->sequence) / sizeof(KzGestureMotion) - 1;
+		= G_N_ELEMENTS(gesture->sequence) - 1;
 	gesture->threshold = 16;
 
 	kz_gesture_reset(gesture); /* init other members */
 
-	gesture->table      = NULL;
-	gesture->table_rows = 0;
+	gesture->items = NULL;
+}
+
+static void
+kz_gesture_dispose (GObject *object)
+{
+	KzGesture *gesture;
+
+	g_return_if_fail(KZ_IS_GESTURE(object));
+
+	gesture = KZ_GESTURE(object);
+
+	kz_gesture_items_unref(gesture->items);
+	gesture->items = NULL;
+
+	if (G_OBJECT_CLASS (parent_class)->dispose) {
+		G_OBJECT_CLASS (parent_class)->dispose(object);
+	}
 }
 
 KzGesture *
@@ -87,6 +109,19 @@ kz_gesture_new (void)
 	return gesture;
 }
 
+void
+kz_gesture_set_items (KzGesture *gesture, KzGestureItems *items)
+{
+	g_return_if_fail(KZ_IS_GESTURE(gesture));
+
+	if (gesture->items)
+		kz_gesture_items_unref(gesture->items);
+	if (items)
+		gesture->items = kz_gesture_items_ref(items);
+	else
+		gesture->items = NULL;
+}
+
 static void
 kz_gesture_stack_motion (KzGesture *gesture, KzGestureMotion motion)
 {
@@ -122,11 +157,11 @@ kz_gesture_start (KzGesture *gesture, gint mode, gint x, gint y)
 gboolean
 kz_gesture_perform (KzGesture *gesture)
 {
-	KzGestureDefTable *table;
+	KzGestureItem *item;
 
-	table = kz_gesture_table_search(gesture);
-	if (table)
-		table->func(table->data, NULL);
+	item = kz_gesture_search_matched_item(gesture);
+	if (item)
+		egg_action_activate(item->action);
 
 	kz_gesture_reset(gesture);
 
@@ -241,31 +276,33 @@ kz_gesture_update_position (KzGesture *gesture, gint x, gint y)
 	}
 }
 
-static KzGestureDefTable *
-kz_gesture_table_search (KzGesture *gesture)
+static KzGestureItem *
+kz_gesture_search_matched_item (KzGesture *gesture)
 {
-	gint i, j;
+	GSList *node;
+	gint j;
 
 	g_return_val_if_fail(KZ_IS_GESTURE(gesture), NULL);
 	if (!kz_gesture_is_started(gesture)) return NULL;
 	if (gesture->sequence[0] == 0) return NULL;
+	if (!gesture->items) return NULL;
 
-	for (i = 0; i < gesture->table_rows; i++)
+	for (node = gesture->items->list; node; node = g_slist_next(node))
 	{
-		KzGestureDefTable *table = &gesture->table[i];
+		KzGestureItem *item = node->data;
 
-		if (table->sequence[0] == 0)
-			continue;
+		if (!item) continue;
+		if (item->sequence[0] == 0) continue;
 
-		for (j = 0; gesture->sequence[j] == table->sequence[j]; j++)
+		for (j = 0; gesture->sequence[j] == item->sequence[j]; j++)
 		{
 			if (gesture->sequence[j + 1] == 0 &&
-			    table->sequence[j + 1] == 0)
+			    item->sequence[j + 1] == 0)
 			{
-				return table;
+				return item;
 			}
 			else if (gesture->sequence[j + 1] == 0 ||
-				 table->sequence[j + 1] == 0)
+				 item->sequence[j + 1] == 0)
 			{
 				break;
 			}
@@ -278,13 +315,84 @@ kz_gesture_table_search (KzGesture *gesture)
 const gchar *
 kz_gesture_get_matched_label (KzGesture *gesture)
 {
-	KzGestureDefTable *table;
+	KzGestureItem *item;
 
 	g_return_val_if_fail(KZ_IS_GESTURE(gesture), NULL);
 
-	table = kz_gesture_table_search (gesture);
-	if (table)
-		return table->label;
-	else
-		return NULL;
+	item = kz_gesture_search_matched_item (gesture);
+	if (!item) return NULL;
+	g_return_val_if_fail(item->action, NULL);
+	return item->action->name;
 }
+
+
+/*****************************************************************************
+ *
+ *   KzGestureItemTable
+ *
+ *****************************************************************************/
+KzGestureItems *
+kz_gesture_items_new ()
+{
+	KzGestureItems *items = g_new0(KzGestureItems, 1);
+	items->list = NULL;
+	items->ref_count = 1;
+	return items;
+}
+
+KzGestureItems *
+kz_gesture_items_ref (KzGestureItems *items)
+{
+	g_return_val_if_fail(items, NULL);
+	items->ref_count++;
+	return items;
+}
+
+void
+kz_gesture_items_unref (KzGestureItems *items)
+{
+	g_return_if_fail(items);
+	items->ref_count--;
+
+	if (items->ref_count == 0) {
+		GSList *node = items->list;
+
+		for(; node; node = g_slist_next(node)) {
+			KzGestureItem *item = node->data;
+
+			if (!item) continue;
+
+			g_object_unref(item->action);
+			item->action = NULL;
+			g_free(item->sequence);
+			item->sequence = NULL;
+			g_free(item);
+		}
+
+		g_slist_free(items->list);
+		items->list = NULL;
+		g_free(items);
+	}
+}
+
+void
+kz_gesture_items_add_action (KzGestureItems *items,
+			     EggAction *action, gint mode,
+			     const KzGestureMotion *sequence)
+{
+	KzGestureItem *item;
+	gint len;
+
+	g_return_if_fail(items);
+	g_return_if_fail(EGG_IS_ACTION(action));
+	g_return_if_fail(sequence && *sequence != KZ_GESTURE_MOTION_TERMINATOR);
+
+	for (len = 0; sequence[len] != KZ_GESTURE_MOTION_TERMINATOR; len++);
+	item = g_new0(KzGestureItem, 1);
+	item->action   = g_object_ref(action);
+	item->mode     = mode;
+	item->sequence = g_memdup(sequence, sizeof(KzGestureMotion) * (len + 1));
+
+	items->list = g_slist_append(items->list, item);
+}
+
