@@ -17,6 +17,8 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
+#include <stdlib.h>
+#include <string.h>
 #include <gtk/gtk.h>
 #include <fcntl.h>
 #include <sys/utsname.h>
